/*	
 *		ELF Code Reader -- Dump2Code v0.30
 *
 *				coded by newbug[at]chroot.org
 */

#include <elf.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>

#ifdef  LIBDISASM
#include "libdis.h"	/* libdisasm */
#endif

#define	DEFAULT_SECTION		".text"
#define	LINE			14
#define	LINE_SIZE		20
#define	SYNTAX_NUM		3

int shnum,phnum;
unsigned char *start;
void err_exit(const char *msg)
{
	perror(msg);
	exit(1);
}

Elf32_Shdr
*save_shdr(unsigned char *start, Elf32_Half size)
{       
        unsigned char *ret;

        ret = (unsigned char *)malloc(shnum*size);
        memcpy((void *)ret, (void *)start, shnum*size);
        return (Elf32_Shdr *)ret;
}

int shsize,shoff;
Elf32_Shdr *
search_shname(Elf32_Shdr *shdr, char *name, char *tab)
{
	Elf32_Shdr *p;
	int i,num;
	
	p = shdr;

	if(name == NULL)
		name = DEFAULT_SECTION;
	/* st_shndx */
	num = atoi(name);
	if(num)
		return p+num;
	/* shname */
	for(i=0;i<shnum;i++)
		if(!strcmp(tab+(p+i)->sh_name, name))
		{
			printf("[%d]%s\n", i, tab+(p+i)->sh_name);
			return p+i;
		}
	return NULL;
}

Elf32_Sym *
search_syname(Elf32_Shdr *symbol, char *name, unsigned char *tab)
{
	int len;
	Elf32_Sym *p;

	for(len=0;len<symbol->sh_size;len+=sizeof(Elf32_Sym))
	{
		p = start+symbol->sh_offset+len;
		if(!strncmp(name, tab+p->st_name,strlen(name)))
			return p;
	}
	return NULL;
}

int is_elf_file(unsigned char *p)
{       
        if(memcmp((void *)p, ELFMAG, strlen(ELFMAG)))
                return 0;
        return 1;
}       



FILE *fp=NULL;
int len=-1, offset=0;
int zero = 0; // zero flag
#ifdef	LIBDISASM
int disas = 0; // disas flag
int syntax = ATT_SYNTAX;
#endif

char *code;

void dumpcode()
{
	int i;
#ifdef	LIBDISASM
	int n,size;
	struct instr curr_inst;
	char SYNTAX[SYNTAX_NUM][15]={"Native SYNTAX",
			 "Intel SYNTAX",
			 "AT&T SYNTAX"};
	
	fprintf(fp, "/* %s */\n",SYNTAX[syntax]);
#endif

	fprintf(fp, "char code[] = \n");
	

	
        code = (char *)malloc(len);
        memcpy((void *)code, (const void *)(start+shoff+offset), len);           
	i=0;
	while(i<len)
	{
#ifdef	LIBDISASM
		disassemble_init(0, syntax);
		if(disas)
		{
			memset((void *)&curr_inst, '\0', sizeof(curr_inst));
			size = disassemble_address(code + i, &curr_inst);
			if(size)
			{
				fprintf(fp, "\t\"");
				for(n=0;n<size;n++)
					fprintf(fp, "\\x%.2x",*(code+i+n)&0xff);
				fprintf(fp, "\"\t\t/* ");
				fprintf(fp, "%s",curr_inst.mnemonic);
				if (curr_inst.dest[0] != 0)
					fprintf(fp, " %s", curr_inst.dest);
				if (curr_inst.src[0] != 0)
					fprintf(fp, ",%s", curr_inst.src);
				if (curr_inst.aux[0] != 0)
					fprintf(fp, ",%s", curr_inst.aux);
				fprintf(fp, " */\n");
				i+=size;
			}
			else
			{
				fprintf(fp,"\"\\x%.2x\"\t/* unknown inst */\n",\
						(*(code+i))&0xff);
				i++;
			}
		}
		else  /* disas == 0 */
		{
#endif	
			static int first = 0;

			if(!first)
			{
				fprintf(fp,"\t\"");
				first = 1;
			}
			/* detect zero byte, break. */
			if(zero && (*(code+i) == 0x0))
        			break;
	                fprintf(fp, "\\x%.2x",(*(code+i))&0xff);
	                if(!((i+1)%LINE))
               		        fprintf(fp, "\"\n\t\"");
			i++;
#ifdef	LIBDISASM
		}
		disassemble_cleanup();
#endif
	}
#ifdef	LIBDISASM
	if(!disas)
#endif		
		fprintf(fp, "\"\n");
	fprintf(fp, ";\n");
	
	free(code);

}


void usage(char *arg)
{
	printf("dump2code v0.30  coded by newbug[at]chroot.org\n");
	printf("Usage: %s ",arg);
	printf("-f filename [-O output filename] [-s section name]\n");
	printf("\t[-S symbol name] [-l length] [-o offset] [-0] [-h]\n");
	#ifdef	LIBDISASM
		printf("\t[-d]\n");
	#endif
	printf("\tdefault_section = \".text\"\n");
	exit(0);
}
void main(int argc,char **argv)
{
	int fd;
	struct stat buf;
	Elf32_Ehdr ehdr;
	Elf32_Shdr *shdr,*rsh,*tab,*symtab; // rsh ==> result section
					    // tab ==> strtab section
	Elf32_Sym *sym;
	unsigned char *shstrtab,*strtab;    // strtab ==> strtab string index
	char ch;
	char *fname=NULL, *shname=NULL, *syname=NULL;
	char s[10]={'\0'};
	int sym_f=0;
	
	if(argc == 1)
		usage(argv[0]);
#ifdef	LIBDISASM	
	while((ch = getopt(argc, argv, "f:l:s:S:o:O:t:0dh"))!=-1)
#else
	while((ch = getopt(argc, argv, "f:l:s:S:o:O:0h"))!=-1)
#endif
	{
		switch(ch)
		{
			case 'f':  /* file to disas */
				fname = (char *)malloc(strlen(optarg));
				strcpy(fname, optarg);
				break;
			case 'O':  /* output file */
				fp = fopen(optarg, "w");
				break;
			case 's':  /* section to disas */
				shname = (char *)malloc(strlen(optarg));
				strcpy(shname, optarg);
				break;
			case 'S':  /* symbol to disas */
				syname = (char *)malloc(strlen(optarg));
				strcpy(syname, optarg);
				sym_f = 1;
				break;
			case 'l':  /* length */
				len = atoi(optarg);
				break;
			case 'o':  /* offset */
				offset = atoi(optarg);
				break;
			case '0':  /* stop when detect zero */
				zero = 1;
				break;	
#ifdef	LIBDISASM
			case 'd':  /* enable disas */
				disas = 1;
				break;
			case 't':
				syntax = atoi(optarg)%3;
				break;
#endif
			case 'h':
			default:
				usage(argv[0]);
		}
	}
	fd = open(fname, O_RDONLY);
	if( fd == -1)
		err_exit("open");
	fstat(fd, &buf);
	start = (unsigned char *)\
		mmap(0, buf.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
	close(fd);
	memcpy((void *)&ehdr, start, sizeof(ehdr));
	if(is_elf_file((unsigned char *)&ehdr))
	{
		shnum = ehdr.e_shnum;
		phnum = ehdr.e_phnum;
		/* Save Section header */
		shdr = (Elf32_Shdr *)save_shdr((unsigned char *) \
					start+ehdr.e_shoff,
			    		ehdr.e_shentsize);
		
		shstrtab = start+(shdr + ehdr.e_shstrndx)->sh_offset;

		if(fp == NULL)
			fp = stdout;
		if(sym_f == 0)	/* dump section */
		{
			rsh = search_shname(shdr, shname, shstrtab);
			if(rsh == NULL)
			{
				printf("Section %s not found.\n",shname);
				goto out;
			}
			if(len == -1)
				len = rsh->sh_size;
			shoff = rsh->sh_offset;
		}
		else	/* dump symbol */
		{
			symtab = search_shname(shdr,".symtab",shstrtab);
			if(symtab == NULL)
			{
				printf("maybe stripped, out.\n");
				goto out;
			}
			tab = search_shname(shdr,".strtab",shstrtab);
			strtab = start + tab->sh_offset;
			sym = search_syname(symtab, syname, strtab);
			if(sym == NULL)
			{
				printf("Symbol %s not found.\n");
				goto out;
			}
			if(!sym->st_size)
			{
				printf("Symbol %s size is 0.\n");
				goto out;
			}
			sprintf(s, "%d",sym->st_shndx);
			rsh = search_shname(shdr, shname, shstrtab);
                        if(len == -1)
                                len = sym->st_size;
	                shoff = sym->st_value-rsh->sh_addr+rsh->sh_offset;
		}
		dumpcode();

out:
		fclose(fp);
		free(shdr);
	}
	else
		printf("Not ELF ...\n");

	free(fname);
	free(shname);
	if(syname)
		free(syname);

	munmap(start, buf.st_size);

}

